#include <stdio.h>
#include <string.h>

#include <alleycat.h>

#include "drupn-addon.h"

#include "drup/lib.h"
#include "drup/types.h"

#include "private/util.h"

/* @throws
 */
#define check_cleaned_up_throw(env) do { \
  if (cleaned_up) throw (env, \
    "Invalid state: already called cleanup() ?" \
  ); \
} while (0);

static void *table_get_num_entries (napi_env env, napi_value table, uint32_t *len) {
  napi_value res;
  napi_call (env, napi_get_property_names (env, table, &res));
  napi_call (env, napi_get_array_length (env, res, len));
  return NULL;
}

static void *get_type (napi_env env, napi_value val, char **ret) {
  napi_valuetype t;
  napi_call (env, napi_typeof (env, val, &t));
  if (t == napi_object) *ret = "object";
  else if (t == napi_undefined) *ret = "undefined";
  else if (t == napi_null) *ret = "null";
  else if (t == napi_boolean) *ret = "boolean";
  else if (t == napi_number) *ret = "number";
  else if (t == napi_string) *ret = "string";
  else if (t == napi_symbol) *ret = "symbol";
  else if (t == napi_object) *ret = "object";
  else if (t == napi_function) *ret = "function";
  else if (t == napi_external) *ret = "external";
  else if (t == napi_bigint) *ret = "bigint";
  // --- @todo this is valid but generates a compiler warning.
  else napi_fatal_error ("get_type ()", NAPI_AUTO_LENGTH, "unknown type", NAPI_AUTO_LENGTH);
  return NULL;
}

static napi_value drupn_authenticate (napi_env env, napi_callback_info info);
static napi_value drupn_cleanup (napi_env env, napi_callback_info info);
static napi_value drupn_generate_secret (napi_env env, napi_callback_info info);
static napi_value drupn_init (napi_env env, napi_callback_info info);
static napi_value drupn_search (napi_env env, napi_callback_info info);
static napi_value drupn_search_init (napi_env env, napi_callback_info info);

static bool cleaned_up = false;

/* @throws
 */
static napi_value drupn_authenticate (napi_env const env, napi_callback_info const info) {
  struct drup_state *state;

  napi_value argv[3];
  uint8_t bus_number;
  uint64_t device_id;
  uint8_t secret16[16];

  {
    void *data;
    size_t argc = 3;

    // --- @todo napi_call?
    napi_get_cb_info (env, info, &argc, argv, NULL, &data);
    if (argc != 3) throw (env,
      "authenticate (): expected exactly %d argument(s) (got %zu)",
      3, argc
    );

    state = (struct drup_state *) data;

    check_cleaned_up_throw (env);
  }

  {
    int32_t bus_number32;
    napi_call (env, napi_get_value_int32 (env, argv[0], &bus_number32))
    bus_number = (uint8_t) bus_number32;
  }

  {
    napi_value nval = argv[1];
    size_t size;
    string_length_latin1 (env, nval, &size)

    if (size != 16) ac_warn ("authenticate (): expected device id to be 16 chars long");

    char device_id_str[17];
    napi_call (env, napi_get_value_string_latin1 (env, nval,
      device_id_str, size + 1, NULL
    ))

    hex_to_uint64 (device_id_str, &device_id);
  }

  {
    napi_typedarray_type type;
    size_t length;
    void *datav;
    uint8_t *data;
    size_t offset;
    napi_call (env, napi_get_typedarray_info (env, argv[2],
      &type, &length, &datav, NULL, &offset
    ))
    if (type != napi_uint8_array) throw (env,
      "authenticate (): `secret` must be a Uint8Array"
    );
    if (length != 16) throw (env,
      "authenticate (): `secret` must be a Uint8Array of length 16"
    );
    data = datav - offset;
    memcpy (secret16, data, 16);
  }

  bool authenticated;
  // --- underlying data array can get garbage collected.
  uint8_t secret16_copy[16];
  memcpy (secret16_copy, secret16, 16);
  bool rc = drup_authenticate (state, bus_number, device_id, secret16_copy, &authenticated);
  if (!rc) throw (env, "Error with authenticate (I/O error?)");

  napi_value nauthenticated;
  napi_value nbauthenticated;
  napi_call (env, napi_create_int32 (env, authenticated, &nauthenticated))
  napi_call (env, napi_coerce_to_bool (env, nauthenticated, &nbauthenticated))
  return nbauthenticated;
}

static napi_value drupn_cleanup (napi_env const env, napi_callback_info const info) {
  struct drup_state *state;
  {
    size_t argc = 0;
    void *data;

    napi_get_cb_info (env, info, &argc, NULL, NULL, &data);
    if (argc != 0) throw (env,
      "cleanup (): expected exactly %d argument(s) (got %zu)",
      0, argc
    );
    state = (struct drup_state *) data;
  }
  if (state) drup_cleanup (state);
  cleaned_up = true;
  return NULL;
}

static napi_value drupn_generate_secret (napi_env const env, napi_callback_info const info) {
  struct drup_state *state;

  napi_value argv[2];
  uint8_t bus_number;
  uint64_t device_id;

  {
    void *data;
    size_t argc = 2;

    napi_get_cb_info (env, info, &argc, argv, NULL, &data);
    if (argc != 2) throw (env,
      "generateSecret (): expected exactly %d argument(s) (got %zu)",
      2, argc
    );

    state = (struct drup_state *) data;

    check_cleaned_up_throw (env);
  }

  {
    int32_t bus_number32;
    napi_call (env, napi_get_value_int32 (env, argv[0], &bus_number32))
    bus_number = (uint8_t) bus_number32;
  }

  {
    napi_value nval = argv[1];
    size_t size;
    string_length_latin1 (env, nval, &size)

    char device_id_str[17];
    if (size != 16) ac_warn ("generateSecret (): expected device id to be 16 chars long");

    napi_call (env, napi_get_value_string_latin1 (env, nval,
      device_id_str, size + 1, NULL
    ))

    hex_to_uint64 (device_id_str, &device_id);
  }

  uint8_t secret[16];
  bool rc = drup_generate_secret (state, bus_number, device_id, secret);
  if (!rc) throw (env, "Error with generateSecret (I/O error?)");

  napi_value nsecret;

  napi_value nsecret_buffer;
  uint8_t *secret_buffer;
  napi_call (env, napi_create_arraybuffer (env,
    16, (void**) &secret_buffer, &nsecret_buffer
  ));
  memcpy (secret_buffer, secret, 16);

  napi_call (env, napi_create_typedarray (env,
    napi_uint8_array, 16, nsecret_buffer, 0, &nsecret
  ));
  return nsecret;
}

/* @throws
 */
static napi_value drupn_init (napi_env const env, napi_callback_info const info) {
  napi_value device_table;

  const struct drup_state *state;

  char *i2c_dev;
  int32_t address;
  ds2482_type subtype;

  {
    size_t argc = 1;
    napi_value argv[1];

    napi_get_cb_info (env, info, &argc, argv, NULL, NULL);
    if (argc != 1) throw (env,
      "init (): expected exactly %d argument(s) (got %zu)",
      1, argc
    );

    device_table = argv[0];
  }

  {
    char *s = NULL;
    get_type (env, device_table, &s);
    if (strcmp(s, "object")) throw (env,
      "init (): expected object (got %s)", s
    );
  }

  {
    napi_value nval;
    napi_call (env, napi_get_named_property (env, device_table, "i2cDev", &nval))

    if (is_undefined (env, nval)) throw (env,
      "init (): missing param “i2cDev”"
    );

    size_t size;
    size_t ret_size;

    string_length_latin1 (env, nval, &size)

    try_malloc_size_nodecl_rnull (i2c_dev, sizeof (char) * (size + 1))
    if (i2c_dev == NULL) throw (env, "Unable to init (oom)");

    napi_call (env, napi_get_value_string_latin1 (env,
      nval, i2c_dev, size + 1, &ret_size
    ))
  }

  {
    napi_value nval;
    napi_call (env, napi_get_named_property (env, device_table, "address", &nval))

    if (is_undefined (env, nval)) throw (env,
      "init (): missing param “address”"
    );

    napi_call (env, napi_get_value_int32 (env, nval, &address))
  }

  {
    napi_value nval;
    napi_call (env, napi_get_named_property (env, device_table, "subtype", &nval))

    if (is_undefined (env, nval)) throw (env,
      "init (): missing param “subtype”"
    );

    size_t ret_size;

    char buf[4];

    napi_call (env, napi_get_value_string_latin1 (env,
      nval, buf, (size_t) 5, &ret_size
    ))

    if (!strcmp (buf, "100")) subtype = DS2482_TYPE_100;
    else if (!strcmp (buf, "800")) subtype = DS2482_TYPE_800;
    else throw (env, "init (): bad subtype: must be '100' or '800'");
  }

  uint32_t l;
  table_get_num_entries (env, device_table, &l);
  if (l != 3) throw (env, "init (): device table should have exactly 3 entries (got %zu)", l);

  try_malloc (struct drup_config, config)
  if (config == NULL) throw (env, "Unable to init (oom)");
  try_malloc (struct drup_device, device)
  if (device == NULL) throw (env, "Unable to init (oom)");

  device->i2c_dev = i2c_dev;
  device->address = (uint8_t) address;
  device->subtype = subtype;

  config->device = device;
  config->verbose = false;

  state = drup_init (config);
  if (state == NULL) throw (env, "Unable to init");

  napi_value nret;
  napi_value nsearch_init;
  napi_value nsearch;
  napi_value nauthenticate;
  napi_value ngenerate_secret;
  napi_value ncleanup;
  napi_call (env, napi_create_object (env, &nret))
  napi_call (env, napi_create_function (env,
    "searchInit", NAPI_AUTO_LENGTH, drupn_search_init, (void *) state, &nsearch_init
  ))
  napi_call (env, napi_set_named_property (env, nret, "searchInit", nsearch_init))
  napi_call (env, napi_create_function (env,
    "search", NAPI_AUTO_LENGTH, drupn_search, (void *) state, &nsearch
  ))
  napi_call (env, napi_set_named_property (env, nret, "search", nsearch))
  napi_call (env, napi_create_function (env,
    "authenticate", NAPI_AUTO_LENGTH, drupn_authenticate, (void *) state, &nauthenticate
  ))
  napi_call (env, napi_set_named_property (env, nret, "authenticate", nauthenticate))
  napi_call (env, napi_create_function (env,
    "generateSecret", NAPI_AUTO_LENGTH, drupn_generate_secret, (void *) state, &ngenerate_secret
  ))
  napi_call (env, napi_set_named_property (env, nret, "generateSecret", ngenerate_secret))
  napi_call (env, napi_create_function (env,
    "cleanup", NAPI_AUTO_LENGTH, drupn_cleanup, (void *) state, &ncleanup
  ))
  napi_call (env, napi_set_named_property (env, nret, "cleanup", ncleanup))

  return nret;
}

/* @throws
 */
static napi_value drupn_search (napi_env const env, napi_callback_info const info) {
  struct drup_state *state;
  // --- 0 is ok, means disable filtering.
  uint8_t device_family = 0;

  {
    void *data;

    size_t argc = 1;
    napi_value argv[1];
    napi_get_cb_info (env, info, &argc, argv, NULL, &data);
    if (argc > 1) throw (env,
      "search (): expected exactly %d or %d argument(s) (got %zu)",
      0, 1, argc
    );

    state = (struct drup_state *) data;

    check_cleaned_up_throw (env);

    if (argc) {
      napi_value arg = argv[0];
      napi_value ndevice_family;
      napi_call (env, napi_get_named_property (env, arg, "deviceFamily", &ndevice_family))

      if (is_undefined (env, ndevice_family)) throw (env,
        "search (): missing param “deviceFamily”"
      );

      int32_t val;
      napi_call (env, napi_get_value_int32 (env, ndevice_family, &val))
      device_family = (uint8_t) val;
    }
  }

  char device_id_str[17];
  uint64_t device_id;
  bool rc = drup_search (state, device_family, &device_id, device_id_str);
  if (!rc) throw (env, "Error with search (I/O error?)");

  napi_value n_array;
  int const length = device_id ? 1 : 0;
  napi_call (env, napi_create_array_with_length (env, length, &n_array))
  if (device_id) {
    napi_value n_device_id;
    napi_call (env, napi_create_string_latin1 (env, device_id_str, 16, &n_device_id))
    napi_call (env, napi_set_element (env, n_array, 0, n_device_id))
  }
  return n_array;
}

/* @throws
 */
static napi_value drupn_search_init (napi_env const env, napi_callback_info const info) {
  struct drup_state *state;
  uint8_t bus_number;

  {
    napi_value nbus_number;
    size_t argc = 1;
    napi_value argv[1];

    void *data;

    napi_get_cb_info (env, info, &argc, argv, NULL, &data);
    if (argc != 1) throw (env,
      "search_init (): expected exactly %d argument(s) (got %zu)",
      1, argc
    );

    state = (struct drup_state *) data;

    check_cleaned_up_throw (env);

    nbus_number = argv[0];
    int32_t val;
    napi_call (env, napi_get_value_int32 (env, nbus_number, &val))
    bus_number = val;
  }

  bool presence = false;
  bool rc = drup_search_init (state, bus_number, &presence);
  if (!rc) throw (env, "Error with search_init (I/O error?)");
  napi_value npresence_int;
  napi_value npresence;
  napi_call (env, napi_create_int32 (env, presence, &npresence_int))
  napi_call (env, napi_coerce_to_bool (env, npresence_int, &npresence))
  return npresence;
}

napi_value create_drupn_addon (napi_env const env) {
  napi_value result;
  napi_call (env, napi_create_object (env, &result))

  napi_value n_init;

  napi_call (env, napi_create_function (env,
    "init", NAPI_AUTO_LENGTH, drupn_init, NULL, &n_init
  ))

  napi_call (env, napi_set_named_property (env,
    result, "init", n_init
  ))

  return result;
}
